"""RSS blueprint for ArchivePodcast."""

from http import HTTPStatus
from pathlib import Path

from flask import Blueprint, Response, current_app, render_template
from lxml import etree

from archivepodcast.instances.config import get_ap_config
from archivepodcast.instances.podcast_archiver import (
    get_about_page_exists,
    get_ap,
)
from archivepodcast.utils.logger import get_logger

logger = get_logger(__name__)
bp = Blueprint("rss", __name__)


@bp.route("/rss/<string:feed>", methods=["GET"])  # type: ignore[untyped-decorator]
def rss(feed: str) -> Response:
    """Send RSS Feed."""
    ap = get_ap()

    ap_conf = get_ap_config()

    logger.debug("Sending rss feed: %s", feed)
    rss_bytes = b""
    rss_str = ""
    try:
        rss_bytes = ap.get_rss_feed(feed)
        rss_str = rss_bytes.decode("utf-8")
    except TypeError:
        return_code = HTTPStatus.INTERNAL_SERVER_ERROR
        return Response(
            render_template(
                "error.html.j2",
                error_code=str(return_code),
                error_text="The developer probably messed something up",
                about_page=get_about_page_exists(),
                app_config=ap_conf.app,
                header=ap.renderer.webpages.generate_header("error.html"),
            ),
            status=return_code,
        )

    except KeyError:
        try:
            tree = etree.parse(Path(current_app.instance_path) / "web" / "rss" / feed)
            rss = etree.tostring(
                tree.getroot(),
                encoding="utf-8",
                method="xml",
                xml_declaration=True,
            )
            rss_str = rss.decode("utf-8")
            logger.warning('❗ Feed "%s" not live, sending cached version from disk', feed)

        # The file isn't there due to user error or not being created yet
        except (FileNotFoundError, OSError):
            return_code = HTTPStatus.NOT_FOUND
            return Response(
                render_template(
                    "error.html.j2",
                    error_code=str(return_code),
                    error_text="Feed not found, you know you can copy and paste yeah?",
                    about_page=get_about_page_exists(),
                    app_config=ap_conf.app,
                    podcasts=ap_conf.podcasts,
                    header=ap.renderer.webpages.generate_header("error.html"),
                ),
                status=return_code,
            )

        except:  # noqa: E722 Bare except since this is a catch all to prevent app crash
            return_code = HTTPStatus.INTERNAL_SERVER_ERROR
            return Response(
                render_template(
                    "error.html.j2",
                    error_code=str(return_code),
                    error_text="Feed not loadable, Internal Server Error",
                    about_page=get_about_page_exists(),
                    app_config=ap_conf.app,
                    podcasts=ap_conf.podcasts,
                    header=ap.renderer.webpages.generate_header("error.html"),
                ),
                status=return_code,
            )

    return Response(rss_str, mimetype="application/rss+xml; charset=utf-8", status=HTTPStatus.OK)



"""Blueprint for static files and special routes like robots.txt and favicon.ico."""

from flask import Blueprint, Response

from archivepodcast.instances.podcast_archiver import (
    send_ap_cached_webpage,
)

bp = Blueprint("static", __name__)


@bp.route("/robots.txt")  # type: ignore[untyped-decorator]
def send_robots() -> Response:
    """Serve robots.txt."""
    return send_ap_cached_webpage("robots.txt")


@bp.route("/favicon.ico")  # type: ignore[untyped-decorator]
def favicon() -> Response:
    """Return the favicon."""
    return send_ap_cached_webpage("static/favicon.ico")


@bp.route("/static/<path:path>")  # type: ignore[untyped-decorator]
def send_static(path: str) -> Response:
    """Serve static files."""
    return send_ap_cached_webpage(f"static/{path}")



"""Webpage caching and management."""

from typing import ClassVar

from archivepodcast.instances.health import health


class Webpage:
    """Represents a cached webpage with its metadata."""

    def __init__(self, path: str, mime: str, content: str | bytes) -> None:
        """Initialise the Webpages object."""
        # Mime types that magic doesn't always get right
        if path.endswith(".js"):
            mime = "text/javascript"
        elif path.endswith(".css"):
            mime = "text/css"
        elif path.endswith(".woff2"):
            mime = "font/woff2"

        self.path: str = path
        self.mime: str = mime
        self.content: str | bytes = content


class Webpages:
    """Manages a collection of cached webpages."""

    WEBPAGE_NICE_NAMES: ClassVar[dict[str, str]] = {
        "index.html": "Home",
        "guide.html": "Guide",
        "filelist.html#/content/": "File List",  # Navigate to /content/ by default
        "webplayer.html": "Web Player",
        "about.html": "About",
    }

    def __init__(self) -> None:
        """Initialise the Webpages object."""
        self._webpages: dict[str, Webpage] = {}

    def __len__(self) -> int:
        """Return the length of the webpages."""
        return len(self._webpages)

    def get_list(self) -> dict[str, str]:
        """Return the items of the webpages."""
        item_list = self._webpages.items()

        return_value = {}
        for _, value in item_list:
            return_value[value.path] = value.mime

        return return_value

    def add(self, path: str, mime: str, content: str | bytes) -> None:
        """Add a webpage."""
        self._webpages[path] = Webpage(path=path, mime=mime, content=content)
        health.set_asset(path, mime)

    def get_all_pages(self) -> dict[str, Webpage]:
        """Return the webpages."""
        return self._webpages

    def get_webpage(self, path: str) -> Webpage:
        """Get a webpage."""
        return self._webpages[path]

    def generate_header(self, path: str, *, debug: bool = False) -> str:
        """Get the header for a webpage."""
        reload_a_tag = """
 | <a href="#" onclick="
const originalText = document.getElementById('debug_status').innerHTML;

function resetText() {
    document.getElementById('debug_status').innerHTML = originalText;
}

fetch('/api/reload')
.then(response => response.json())
.then(data => {
    console.log(data.msg);
    document.getElementById('debug_status').innerHTML = 'RELOAD SENT';
    setTimeout(resetText, 3000);
})
.catch(error => console.error(error));
return false;
">Reload</a>"""

        header = "<header>"

        for webpage in self.WEBPAGE_NICE_NAMES:
            if webpage == "about.html":
                about_page_exists = self._webpages.get("about.html") or path == "about.html"
                if not about_page_exists:
                    continue

            if webpage == path:
                header += f'<div class="active">{self.WEBPAGE_NICE_NAMES[webpage]}</div> | '
            else:
                header += f'<a href="{webpage}">{self.WEBPAGE_NICE_NAMES[webpage]}</a> | '

        header = header[:-3]

        if debug:
            if path != "health.html":
                header += ' | <a href="/health">Health</a>'
            else:
                header += ' | <div class="active">Health</div>'
            header += reload_a_tag.replace("\n", "")
            header += ' | <a href="/console" target="_blank">Flask Console</a>'
            header += ' | <a id="debug_status" style="color: #ff0000">DEBUG ENABLED</a>'

        header += "<hr></header>"
        return header



"""Actually Download Assets."""

import contextlib
import shutil
import sys
import time
from pathlib import Path

import aiohttp
import ffmpeg
from aiobotocore.session import get_session
from botocore.exceptions import ClientError as S3ClientError

from archivepodcast.config import AppConfig, PodcastConfig
from archivepodcast.instances.config import get_ap_config_s3_client
from archivepodcast.instances.path_cache import local_file_cache, s3_file_cache
from archivepodcast.instances.path_helper import get_app_paths
from archivepodcast.utils.log_messages import log_aiohttp_exception
from archivepodcast.utils.logger import get_logger
from archivepodcast.utils.s3 import S3File
from archivepodcast.utils.time import warn_if_too_long

from .constants import CONTENT_TYPES, DOWNLOAD_RETRY_COUNT, FFMPEG_INFO
from .helpers import delay_download

logger = get_logger(__name__)


def check_ffmpeg() -> None:
    """Check if ffmpeg is installed."""
    ffmpeg_paths = [
        Path("/usr/bin/ffmpeg"),
        Path("/usr/local/bin/ffmpeg"),
        Path("C:/Program Files/ffmpeg/bin/ffmpeg.exe"),
        Path("C:/ffmpeg/bin/ffmpeg.exe"),
    ]
    found_manually = any(ffmpeg_path.exists() for ffmpeg_path in ffmpeg_paths)

    if not shutil.which("ffmpeg") and not found_manually:
        logger.error(FFMPEG_INFO)
        sys.exit(1)


check_ffmpeg()


class AssetDownloader:
    """Asset Downloader object."""

    def __init__(
        self,
        podcast: PodcastConfig,
        app_config: AppConfig,
        *,
        s3: bool,
        aiohttp_session: aiohttp.ClientSession,
    ) -> None:
        """Initialise the AssetDownloader object."""
        logger.trace("Initialising AssetDownloader for podcast: %s", podcast.name_one_word)
        self._podcast = podcast
        self._app_config = app_config
        self._s3 = s3
        self._aiohttp_session = aiohttp_session
        self._feed_download_healthy: bool = True
        self._rss_file_path = get_app_paths().web_root / "rss" / podcast.name_one_word

    # region Download Methods

    async def _download_asset(self, url: str, title: str, extension: str = "", file_date_string: str = "") -> None:
        """Download asset from url with appropriate file name."""
        spacer = ""
        if file_date_string != "":
            spacer = "-"

        content_dir = get_app_paths().web_root / "content" / self._podcast.name_one_word
        file_path = content_dir / f"{file_date_string}{spacer}{title}{extension}"

        if not await self._check_path_exists(file_path):  # if the asset hasn't already been downloaded
            await self._download_to_local(url, file_path)
            logger.debug("Downloaded asset: %s", file_path)

            # For if we are using s3 as a backend
            # wav logic since this gets called in handle_wav
            if extension != ".wav" and self._s3:
                await self._upload_asset_s3(file_path, extension)

        else:
            logger.trace(f"Already downloaded: {title}{extension}")

    async def _download_to_local(self, url: str, file_path: Path) -> None:
        """Download the asset from the url."""
        logger.debug("[%s] Downloading: %s", self._podcast.name_one_word, url)

        async def _attempt_download() -> bool:
            """Attempt to download the asset."""
            try:
                logger.trace("[%s] Downloading asset from URL: %s", self._podcast.name_one_word, url)
                start_time = time.time()
                async with self._aiohttp_session.get(url) as response:
                    response.raise_for_status()
                    file_path.parent.mkdir(parents=True, exist_ok=True)
                    with file_path.open("wb") as asset_file:
                        while True:
                            chunk = await response.content.read(8192)
                            if not chunk:
                                break
                            asset_file.write(chunk)

                warn_if_too_long(f"download asset: {file_path}", time.time() - start_time, large_file=True)

            except aiohttp.ClientError as e:
                self._feed_download_healthy = False
                log_aiohttp_exception(self._podcast.name_one_word, url, e, logger)
                return False

            logger.info("[%s] Downloaded asset to: %s", self._podcast.name_one_word, file_path)

            return True

        success = False
        for n in range(DOWNLOAD_RETRY_COUNT):
            success = await _attempt_download()
            if success:
                break
            await delay_download(n)
        if not success:
            logger.error("[%s] Failed to download asset after multiple attempts: %s", self._podcast.name_one_word, url)
            return

        logger.debug("[%s] Success, downloaded to %s", self._podcast.name_one_word, file_path)

        if not self._s3:
            self._append_to_local_paths_cache(file_path)

    async def _download_cover_art(
        self,
        url: str,
        title: str,
        extension: str = "",
        *,
        s3_remove_original: bool = True,
    ) -> None:
        """Download cover art from url with appropriate file name."""
        content_dir = get_app_paths().web_root / "content" / self._podcast.name_one_word
        cover_art_destination = content_dir / f"{title}{extension}"

        remote_file_found = False
        local_file_found = self._check_local_path_exists(cover_art_destination)

        # If we are using s3
        #    we haven't found the local file
        #    we will remove the original after upload
        # Check s3 for the file
        if (
            self._s3
            and not local_file_found
            and s3_remove_original
            and await self._check_path_exists(cover_art_destination)
        ):
            remote_file_found = True

        # Download to local if we aren't using s3 and haven't found it locally
        if not self._s3 and not local_file_found:
            await self._download_to_local(url, cover_art_destination)

        # Download to local if we haven't found it locally or remotely
        if self._s3 and not local_file_found and not remote_file_found:
            await self._download_to_local(url, cover_art_destination)

        # Download to local if we are using s3 but not removing original and haven't found it locally
        if self._s3 and not s3_remove_original and not local_file_found:
            await self._download_to_local(url, cover_art_destination)

        # If we (now) have a file here, upload to s3 if needed
        if self._s3 and (local_file_found or not remote_file_found):
            logger.debug(
                "[%s] Uploading podcast cover art to s3 not deleting local file to allow overriding",
                self._podcast.name_one_word,
            )
            await self._upload_asset_s3(cover_art_destination, extension, remove_original=s3_remove_original)

    async def _handle_wav(self, url: str, title: str, extension: str = "", file_date_string: str = "") -> int:
        """Convert podcasts that have wav episodes 😔. Returns new file length."""
        logger.trace("[%s] Handling wav file: %s", self._podcast.name_one_word, title)
        new_length = None
        spacer = ""  # This logic can be removed since WAVs will always have a date
        if file_date_string != "":
            spacer = "-"

        content_dir = get_app_paths().web_root / "content" / self._podcast.name_one_word
        wav_file_path: Path = content_dir / f"{file_date_string}{spacer}{title}.wav"
        mp3_file_path: Path = content_dir / f"{file_date_string}{spacer}{title}.mp3"

        # If we need do download and convert a wav there is a small chance
        # the user has had ffmpeg issues, remove existing files to play it safe
        if wav_file_path.exists():
            with contextlib.suppress(Exception):
                wav_file_path.unlink()
                mp3_file_path.unlink()

        # If the asset hasn't already been downloaded and converted
        if not await self._check_path_exists(mp3_file_path):
            await self._download_asset(
                url,
                title,
                extension,
                file_date_string,
            )

            logger.info("♻ Converting episode %s to mp3", title)
            logger.debug("♻ MP3 File Path: %s", mp3_file_path)

            input_wav = ffmpeg.input(filename=wav_file_path)
            ff = ffmpeg.output(
                input_wav,
                filename=mp3_file_path,
                codec="libmp3lame",
                aq="4",
            )  # VBR v4 might be overkill for voice buy I pick it

            ff.run()

            logger.info("♻ Done")

            # Remove wav since we are done with it
            logger.info("♻ Removing wav version of %s", title)
            if wav_file_path.exists():
                wav_file_path.unlink()
            logger.info("♻ Done")

            if self._s3:
                await self._upload_asset_s3(mp3_file_path, extension)
        else:
            logger.debug("Episode has already been converted: %s", mp3_file_path)

        if self._s3:
            # Convert mp3_file_path to a Path object and make relative to web_root
            s3_file_path = Path(mp3_file_path).relative_to(get_app_paths().web_root)

            # Convert to posix path (forward slashes) for S3
            s3_key = s3_file_path.as_posix()

            msg = f"Checking length of s3 object: {s3_key}"
            logger.trace("[%s] %s", self._podcast.name_one_word, msg)

            session = get_session()
            ap_s3_config = get_ap_config_s3_client()
            async with session.create_client("s3", **ap_s3_config.__dict__) as s3_client:
                response = await s3_client.head_object(Bucket=self._app_config.s3.bucket, Key=s3_key)

            new_length = response["ContentLength"]
            msg = f"Length of converted wav file {s3_key}: {new_length} bytes, stored in s3"
        else:
            new_length = mp3_file_path.stat().st_size
            msg = f"Length of converted wav file: {mp3_file_path} {new_length} bytes, stored locally"

        logger.trace("[%s] %s", self._podcast.name_one_word, msg)

        return new_length

    # region S3 Upload

    async def _upload_asset_s3(self, file_path: Path, extension: str, *, remove_original: bool = True) -> None:
        """Upload asset to s3."""
        if not self._s3:
            logger.error("[%s] s3 client not found, cannot upload", self._podcast.name_one_word)
            return
        content_type = CONTENT_TYPES[extension]
        file_path = Path(file_path)
        if not file_path.is_absolute():
            file_path = get_app_paths().web_root / file_path
        s3_path = file_path.relative_to(get_app_paths().web_root).as_posix()
        s3_path = s3_path.removeprefix("/")

        if not remove_original:
            # So if we are not removing the original, we can check if we can skip the upload
            file_size = file_path.stat().st_size
            if s3_file_cache.check_file_exists(s3_path, file_size):
                logger.debug(
                    "[%s] File: %s exists in s3_paths_cache and matches in size, skipping upload",
                    self._podcast.name_one_word,
                    s3_path,
                )
                return

        try:
            # Upload the file
            session = get_session()
            ap_s3_config = get_ap_config_s3_client()
            async with session.create_client("s3", **ap_s3_config.__dict__) as s3_client:
                if remove_original:
                    logger.info("[%s] Uploading to s3: %s", self._podcast.name_one_word, s3_path)
                else:
                    logger.debug("[%s] Uploading to s3: %s", self._podcast.name_one_word, s3_path)

                start_time = time.time()
                await s3_client.put_object(
                    Bucket=self._app_config.s3.bucket,
                    Key=s3_path,
                    Body=file_path.read_bytes(),
                    ContentType=content_type,
                )
                warn_if_too_long(
                    f"[{self._podcast.name_one_word}] upload asset to s3: {s3_path}",
                    time.time() - start_time,
                    large_file=True,
                )
                logger.trace("[%s] Uploaded asset to s3: %s", self._podcast.name_one_word, s3_path)

            s3_file_cache.add_file(S3File(key=s3_path, size=file_path.stat().st_size))

            if remove_original:
                logger.info("[%s] Removing local file: %s", self._podcast.name_one_word, file_path)
                try:
                    Path(file_path).unlink()
                except FileNotFoundError:  # Some weirdness when in debug mode, otherwise i'd use contextlib.suppress
                    msg = f"Could not remove the local file, the source file was not found: {file_path}"
                    logger.exception("[%s] %s", self._podcast.name_one_word, msg)

        except FileNotFoundError:
            self._feed_download_healthy = False
            logger.exception(
                "[%s] Could not upload to s3, the source file was not found: %s", self._podcast.name_one_word, file_path
            )
        except Exception:
            self._feed_download_healthy = False
            logger.exception("[%s] Unhandled s3 error: %s", self._podcast.name_one_word, file_path)

    # region Helpers

    def _check_local_path_exists(self, file_path: Path) -> bool:
        """Check if the file exists locally."""
        file_exists = file_path.is_file()

        if file_exists:
            self._append_to_local_paths_cache(file_path)
            logger.trace("File: %s exists locally", file_path)
        else:
            logger.trace("File: %s does not exist locally", file_path)

        return file_exists

    async def _check_path_exists(self, file_path: Path | str) -> bool:
        """Check the path, s3 or local."""
        file_exists = False

        if self._s3:
            # Convert file_path to a Path object if it isn't already
            file_path = Path(file_path)

            # If it's an absolute path and under web_root, make it relative to web_root
            if file_path.is_absolute() and file_path.is_relative_to(get_app_paths().web_root):
                file_path = file_path.relative_to(get_app_paths().web_root)

            # Convert to a posix path (forward slashes) and ensure no leading slash
            s3_key = file_path.as_posix().lstrip("/")

            if not s3_file_cache.check_file_exists(s3_key):
                session = get_session()
                ap_s3_config = get_ap_config_s3_client()

                async with session.create_client("s3", **ap_s3_config.__dict__) as s3_client:
                    try:
                        # Head object to check if file exists
                        my_object = await s3_client.head_object(Bucket=self._app_config.s3.bucket, Key=s3_key)
                        logger.debug(
                            "File: %s exists in s3 bucket",
                            s3_key,
                        )
                        s3_file_cache.add_file(S3File(key=s3_key, size=my_object.get("ContentLength", 0)))
                        file_exists = True

                    except S3ClientError as e:
                        if e.response.get("Error", {}).get("Code") == "404":
                            logger.debug(
                                "File: %s does not exist 🙅‍ in the s3 bucket",
                                s3_key,
                            )
                        else:
                            logger.exception("s3 check file exists errored out?")
                    except Exception:  # pylint: disable=broad-exception-caught
                        logger.exception("Unhandled s3 Error:")

            else:
                logger.trace("s3 path %s exists in s3_paths_cache, skipping", s3_key)
                file_exists = True

        else:
            if not isinstance(file_path, Path):
                file_path = Path(file_path)
            file_exists = self._check_local_path_exists(file_path)

        return file_exists

    def _append_to_local_paths_cache(self, file_path: Path) -> None:
        file_path = Path(file_path).relative_to(get_app_paths().web_root)

        if not local_file_cache.check_exists(file_path):
            local_file_cache.add_file(file_path)



"""Blueprint for Content Serving and RSS Feeds."""

from http import HTTPStatus
from pathlib import Path

from flask import Blueprint, Response, current_app, redirect, send_from_directory
from werkzeug.wrappers.response import Response as WerkzeugResponse

from archivepodcast.instances.config import get_ap_config
from archivepodcast.instances.path_helper import get_app_paths
from archivepodcast.utils.logger import get_logger

logger = get_logger(__name__)
bp = Blueprint("content", __name__)


@bp.route("/content/<path:path>")  # type: ignore[untyped-decorator]
def send_content(path: str) -> Response | WerkzeugResponse:
    """Serve Content."""
    ap_conf = get_ap_config()

    if ap_conf.app.storage_backend == "s3":
        path_obj = Path(path)
        web_root = Path(get_app_paths().web_root)
        relative_path = str(path_obj).replace(str(web_root), "")  # The easiest way to get the "relative" path
        new_path = ap_conf.app.s3.cdn_domain.encoded_string() + "content/" + relative_path
        return redirect(location=new_path, code=HTTPStatus.TEMPORARY_REDIRECT)

    web_dir = Path(current_app.instance_path) / "web" / "content"
    return send_from_directory(str(web_dir), path)



"""Module for local file caching functionality."""

from pathlib import Path


class LocalFileCache:
    """Class representing a local file cache."""

    def __init__(self) -> None:
        """Initialise the local file cache."""
        self._files: list[Path] | None = None

    def refresh(self, web_root: Path) -> None:
        """Refresh the local file cache."""
        self._files = [path.relative_to(web_root) for path in web_root.rglob("*") if path.is_file()]
        self._files.sort()

    def get_all(self) -> list[Path]:
        """Get all cached file paths."""
        if self._files is None:
            msg = "File cache is not initialized. Call refresh() first."
            raise ValueError(msg)
        return self._files

    def get_all_str(self) -> list[str]:
        """Get all cached file paths as strings."""
        return [str(path) for path in self.get_all()]

    def check_exists(self, file_path: Path) -> bool:
        """Check if a file path exists in the cache."""
        return file_path in self.get_all()

    def add_file(self, file_path: Path) -> None:
        """Add a new file path to the cache."""
        if self._files is None:
            msg = "File cache is not initialized. Call refresh() first."
            raise ValueError(msg)
        if file_path not in self._files:
            self._files.append(file_path)
            self._files.sort()



"""Flask web application for archiving and serving podcasts."""

import asyncio
import time
from pathlib import Path

from flask import Flask, Response
from rich.traceback import install

from .archiver import PodcastArchiver
from .blueprints import bp_api, bp_content, bp_rss, bp_static, bp_webpages
from .instances import podcast_archiver
from .instances.config import get_ap_config
from .instances.health import health
from .instances.path_helper import get_app_paths
from .instances.profiler import event_times
from .utils import logger as ap_logger
from .utils.log_messages import log_intro
from .utils.profiler import get_event_times_str
from .version import __version__

__all__ = ["__version__", "create_app", "run_ap_adhoc"]

# Don't if we are in lambda
if not ap_logger.force_simple_logger():
    install()


def create_app(instance_path_override: str | None = None) -> Flask:
    """Create and configure the Flask application instance."""
    start_time = time.time()

    app = Flask(
        __name__,
        instance_relative_config=True,
        static_folder=None,
        instance_path=instance_path_override,
    )  # Create Flask app object

    ap_conf = get_ap_config(Path(app.instance_path) / "config.json")

    if ap_conf.flask.TESTING and not app.instance_path.startswith("/tmp"):  # noqa: S108
        msg = "Flask TESTING mode requires instance_path to be a tmp_path."
        raise ValueError(msg)

    ap_conf.write_config(Path(app.instance_path) / "config.json")

    ap_logger.setup_logger(app, ap_conf.logging)  # Setup logger with config

    app.logger.info("Instance path is: %s", app.instance_path)

    # Flask config, at the root of the config object.
    app.config.from_object(ap_conf.flask)

    app.register_blueprint(bp_api)
    app.register_blueprint(bp_content)
    app.register_blueprint(bp_rss)
    app.register_blueprint(bp_static)
    app.register_blueprint(bp_webpages)

    # For modules that need information from the app object we need to start them under `with app.app_context():`
    # Since in the blueprint_one module, we use `from flask import current_app` to get the app object to get the config
    with app.app_context():
        podcast_archiver.initialise_archivepodcast()

    app.app_context().push()  # God knows what does this does but it fixes everything

    @app.errorhandler(404)  # type: ignore[untyped-decorator]
    def invalid_route(e: str) -> Response:
        """404 Handler."""
        app.logger.debug("Error handler: invalid_route: %s", e)
        return podcast_archiver.generate_404()

    duration = time.time() - start_time
    log_intro("webapp", app.logger)
    event_times.set_event_time("create_app", duration)
    app.logger.info("Starting Web Server: %s", ap_conf.app.inet_path)

    return app


def run_ap_adhoc(
    instance_path: Path,
) -> None:
    """Main for adhoc running."""
    logger = ap_logger.get_logger(__name__)

    start_time = time.time()

    config_path = instance_path / "config.json"

    ap_conf = get_ap_config(config_path=config_path)
    ap_conf.write_config(config_path)

    ap_logger.setup_logger(app=None, logging_conf=ap_conf.logging)  # Setup logger with config

    podcast_archiver_start_time = time.time()

    get_app_paths(root_path=Path.cwd(), instance_path=instance_path)

    ap = PodcastArchiver(
        app_config=ap_conf.app,
        podcast_list=ap_conf.podcasts,
        debug=False,  # The debug of the ap object is only for the Flask web server
    )
    event_times.set_event_time("PodcastArchiver", time.time() - podcast_archiver_start_time)

    ap.grab_podcasts()
    asyncio.run(ap.write_health_s3())
    event_times.set_event_time("/", time.time() - start_time)

    logger.trace(health.get_health().model_dump_json(indent=4))
    logger.trace(event_times.model_dump_json(indent=4))
    logger.info(get_event_times_str(event_times))
    logger.info("Done!")



"""Instances for ArchivePodcast application."""

from pathlib import Path

from pydantic import BaseModel

from archivepodcast.config import ArchivePodcastConfig
from archivepodcast.utils.logger import get_logger

logger = get_logger(__name__)

_conf_cache: ArchivePodcastConfig | None = None


def get_ap_config(config_path: Path | None = None) -> ArchivePodcastConfig:
    """Get the global ArchivePodcastConfig instance."""
    global _conf_cache  # noqa: PLW0603
    if _conf_cache is None:
        from archivepodcast.config import ArchivePodcastConfig  # noqa: PLC0415

        if config_path is None:
            msg = "config_path must be provided the first time get_ap_config is called"
            raise ValueError(msg)

        _conf_cache = ArchivePodcastConfig().force_load_config_file(config_path)

    return _conf_cache


class S3ClientConfig(BaseModel):
    """Configuration for S3 client."""

    aws_secret_access_key: str
    aws_access_key_id: str
    region_name: str | None = None
    endpoint_url: str | None


def get_ap_config_s3_client() -> S3ClientConfig:
    """Get the S3 client from the global ArchivePodcastConfig instance."""
    ap_config = get_ap_config()

    return S3ClientConfig(
        aws_secret_access_key=ap_config.app.s3.secret_access_key,
        aws_access_key_id=ap_config.app.s3.access_key_id,
        region_name=ap_config.app.s3.region if ap_config.app.s3.region else None,
        endpoint_url=ap_config.app.s3.api_url.encoded_string() if ap_config.app.s3.api_url else None,
    )



"""Module to handle the ArchivePodcast object."""

import asyncio
import contextlib
import time
from pathlib import Path
from typing import TYPE_CHECKING

import aiohttp
from aiobotocore.session import get_session
from lxml import etree
from pydantic import BaseModel

from archivepodcast.downloader import PodcastsDownloader
from archivepodcast.downloader.constants import USER_AGENT
from archivepodcast.instances.config import get_ap_config_s3_client
from archivepodcast.instances.health import health
from archivepodcast.instances.path_cache import local_file_cache, s3_file_cache
from archivepodcast.instances.path_helper import get_app_paths
from archivepodcast.instances.profiler import event_times
from archivepodcast.utils.logger import get_logger
from archivepodcast.utils.rss import tree_no_episodes
from archivepodcast.utils.time import warn_if_too_long

from .webpage_renderer import WebpageRenderer

if TYPE_CHECKING:
    from archivepodcast.config import AppConfig, PodcastConfig  # pragma: no cover
else:
    AppConfig = object
    PodcastConfig = object

logger = get_logger(__name__)


class APFileList(BaseModel):
    """Podcast file list response model."""

    base_url: str
    files: list[str]


class AIOHttpClientSessionHelper:
    """Helper class to manage a single aiohttp ClientSession."""

    session: aiohttp.ClientSession | None = None

    def get_session(self) -> aiohttp.ClientSession:
        """Get or create the aiohttp ClientSession."""
        if self.session is None or self.session.closed:
            self.session = aiohttp.ClientSession(
                # timeout, 5 minutes, why not
                timeout=aiohttp.ClientTimeout(),
                # tcp connector, 100 is default connection limit, force_close seems to fix some issues
                connector=aiohttp.TCPConnector(),
                raise_for_status=True,
                headers={"User-Agent": USER_AGENT},
            )
        return self.session

    async def close_session(self) -> None:
        """Close the aiohttp ClientSession."""
        if self.session is not None and not self.session.closed:
            await self.session.close()


aiohttp_client_helper = AIOHttpClientSessionHelper()


class PodcastArchiver:
    """Main podcast archiving system that coordinates downloading, storage and serving of podcasts."""

    # region Init
    def __init__(
        self,
        app_config: AppConfig,
        podcast_list: list[PodcastConfig],
        *,
        debug: bool = False,
    ) -> None:
        """Initialise the ArchivePodcast object."""
        self.debug = debug

        # Health object
        health.update_core_status(currently_loading_config=True)
        health.set_host_info(app_config)

        self.s3: bool = app_config.storage_backend == "s3"
        self.renderer = WebpageRenderer(app_config=app_config, podcast_list=podcast_list, s3=self.s3, debug=debug)

        # Set the config and podcast list
        self._app_config: AppConfig = app_config
        self.podcast_list: list[PodcastConfig] = podcast_list
        self.podcast_rss: dict[str, bytes] = {}

        if self.s3:
            logger.info("Using s3 as storage backend, bucket: %s", app_config.s3.bucket)

        self.load_config(app_config, podcast_list)

        # Done, update health
        health.update_core_status(currently_loading_config=False)

    def load_config(self, app_config: AppConfig, podcast_list: list[PodcastConfig]) -> None:
        """Load the config from the config file."""
        self._app_config = app_config
        self.podcast_list = podcast_list
        self._make_folder_structure()

    # endregion

    def get_rss_feed(self, feed: str) -> bytes:
        """Return the rss file for a given feed."""
        return self.podcast_rss[feed]

    # region Archive

    def grab_podcasts(self) -> None:
        """Download and process all configured podcasts.

        Updates health metrics and regenerates file listings after processing.
        """
        grab_podcasts_start_time = time.time()
        health.update_core_status(last_run=int(grab_podcasts_start_time))
        event_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(event_loop)

        # If you want to see whats going on in the event loop, uncomment this and indent until close()
        # When the program is running use the commands
        # $ python -m aiomonitor.cli
        # > ps
        # st <task id>

        # import aiomonitor  # noqa: ERA001
        # with aiomonitor.start_monitor(event_loop):
        # Part 1: Update the file cache to know what files we have already downloaded
        event_loop.run_until_complete(self.update_file_cache())
        event_times.set_event_time("grab_podcasts/Update file cache", time.time() - grab_podcasts_start_time)

        # Part 2: Download and process all podcasts
        ## Event Loop
        podcast_start_time = time.time()

        ## Create Task List
        podcast_tasks = []
        for podcast in self.podcast_list:
            task = self._grab_podcast_with_metrics(podcast)
            podcast_tasks.append(task)

        podcast_tasks.append(self.renderer.render_files())

        ## Run Tasks
        event_loop.run_until_complete(asyncio.gather(*podcast_tasks))
        event_times.set_event_time("grab_podcasts/Scrape", time.time() - podcast_start_time)

        # Part 3: Render files and cleanup
        ## Event Loop
        cleanup_start_time = time.time()

        ap_file_list = event_loop.run_until_complete(self.get_file_list())

        ## Create Task List
        cleanup_tasks = []
        cleanup_tasks.append(self.renderer.render_filelist_html(ap_file_list))
        cleanup_tasks.append(aiohttp_client_helper.close_session())

        ## Run Tasks
        event_loop.run_until_complete(asyncio.gather(*cleanup_tasks))
        event_loop.close()
        event_times.set_event_time("grab_podcasts/Post Scrape", time.time() - cleanup_start_time)

        # Final timing
        total_duration = time.time() - grab_podcasts_start_time
        event_times.set_event_time("grab_podcasts", total_duration)

    async def _grab_podcast_with_metrics(self, podcast: "PodcastConfig") -> None:
        """Wrapper to handle metrics and error handling for individual podcast processing."""
        logger.trace("Starting _grab_podcast_with_metrics for podcast: %s", podcast.name_one_word)
        podcast_grab_start_time = time.time()

        aiohttp_session = aiohttp_client_helper.get_session()

        try:
            await self._grab_podcast(podcast, aiohttp_session=aiohttp_session)
        except Exception:
            logger.exception("Error grabbing podcast: %s", podcast.name_one_word)
            health.update_podcast_status(podcast.name_one_word, healthy_feed=False)

        elapsed_time = time.time() - podcast_grab_start_time
        event_times.set_event_time(f"grab_podcasts/Scrape/{podcast.name_one_word}", elapsed_time)
        logger.trace("Exiting _grab_podcast_with_metrics for podcast: %s", podcast.name_one_word)

    async def _grab_podcast(self, podcast: PodcastConfig, aiohttp_session: aiohttp.ClientSession) -> None:
        """Function to download a podcast and store the rss."""
        tree = None
        previous_feed = b""
        podcasts_downloader = PodcastsDownloader(
            podcast=podcast,
            app_config=self._app_config,
            s3=self.s3,
            aiohttp_session=aiohttp_session,
        )
        logger.info("[%s] Processing podcast to archive: %s", podcast.name_one_word, podcast.new_name)

        with contextlib.suppress(KeyError):  # Set the previous feed var if it exists
            previous_feed = self.podcast_rss[podcast.name_one_word]

        rss_file_path = get_app_paths().web_root / "rss" / podcast.name_one_word

        if podcast.name_one_word == "":
            logger.error("Podcast has no name_one_word set in config, cannot proceed")
            return

        if podcast.live is True:  # download all the podcasts
            tree = await podcasts_downloader.download_podcast()
            if tree:
                last_fetched = int(time.time())
                health.update_podcast_status(podcast.name_one_word, rss_fetching_live=True, last_fetched=last_fetched)
            else:
                # There should be a previous error message too
                logger.error("Unable to download podcast: %s", podcast.name_one_word)

        else:
            logger.info('[%s] "live": false, in config so not fetching new episodes', podcast.name_one_word)
            health.update_podcast_status(podcast.name_one_word, rss_fetching_live=False)

        if tree_no_episodes(tree):  # If there are no episodes, we can't host it
            tree = None

        if tree is None:  # Serving a podcast that we can't currently download?, load it from file
            tree = self._load_rss_from_file(podcast, rss_file_path)

        if tree_no_episodes(tree):  # If there are still not episodes, we still can't host it
            tree = None

        if tree is not None:
            await self._update_rss_feed(podcast, tree, previous_feed)
            health.update_podcast_episode_info(podcast.name_one_word, tree)

        else:
            logger.error("Unable to host podcast: %s, something is wrong", podcast.name_one_word)
            health.update_podcast_status(podcast.name_one_word, rss_available=False)

        del tree

        logger.trace("Exiting _grab_podcast for %s", podcast.name_one_word)

    # region RSS Feed Handling
    async def _update_rss_feed(
        self,
        podcast: PodcastConfig,
        tree: etree._ElementTree,
        previous_feed: bytes,
    ) -> None:
        """Update the rss feed, in memory and s3."""
        self.podcast_rss.update(
            {
                podcast.name_one_word: etree.tostring(
                    tree.getroot(),
                    encoding="utf-8",
                    method="xml",
                    xml_declaration=True,
                )
            }
        )
        logger.info(
            "[%s] Hosted feed: %srss/%s",
            podcast.name_one_word,
            self._app_config.inet_path,
            podcast.name_one_word,
        )

        # Upload to s3 if we are in s3 mode
        if (
            self.s3
            and previous_feed
            != self.podcast_rss[
                podcast.name_one_word
            ]  # This doesn't work when feed has build dates times on it, patreon for one
        ):
            session = get_session()
            s3_config = get_ap_config_s3_client()

            async with session.create_client("s3", **s3_config.model_dump()) as s3_client:
                try:
                    # Upload the file
                    start_time = time.time()
                    logger.trace("Uploading feed %s to s3...", podcast.name_one_word)
                    await s3_client.put_object(
                        Body=self.podcast_rss[podcast.name_one_word],
                        Bucket=self._app_config.s3.bucket,
                        Key="rss/" + podcast.name_one_word,
                        ContentType="application/rss+xml",
                    )
                    warn_if_too_long(f"upload feed {podcast.name_one_word} to s3", time.time() - start_time)

                    logger.debug("[%s] Uploaded feed to s3", podcast.name_one_word)
                except Exception:  # pylint: disable=broad-exception-caught
                    logger.exception("Unhandled s3 error trying to upload the file: %s")
        health.update_podcast_status(podcast.name_one_word, rss_available=True)
        logger.trace("Exiting _update_rss_feed")

    def _load_rss_from_file(self, podcast: PodcastConfig, rss_file_path: Path) -> etree._ElementTree | None:
        """Load the rss from file."""
        tree = None
        if podcast.live is False:
            logger.info("[%s] Loading rss from file: %s", podcast.name_one_word, rss_file_path)
        else:
            logger.warning("[%s] Loading rss from file: %s", podcast.name_one_word, rss_file_path)
        if rss_file_path.is_file():
            try:
                tree = etree.parse(rss_file_path)
            except etree.XMLSyntaxError:
                logger.exception("Error parsing rss file: %s", rss_file_path)
        elif rss_file_path.is_dir():
            logger.error("Calculated RSS feed path is a directory, not a file: %s", rss_file_path)
        else:
            logger.error("Cannot find rss feed file: %s", rss_file_path)

        return tree

    # region Housekeeping

    def _make_folder_structure(self) -> None:
        """Ensure that web_root folder structure exists."""
        logger.debug("Checking folder structure")

        app_paths = get_app_paths()

        folders = [
            app_paths.instance_path,
            app_paths.web_root,
            app_paths.web_root / "rss",
            app_paths.web_root / "content",
        ]

        folders.extend(app_paths.web_root / "content" / entry.name_one_word for entry in self.podcast_list)
        for folder in folders:
            try:
                folder.mkdir(parents=True, exist_ok=True)
            except PermissionError as exc:
                err = (
                    f"You do not have permission to create folder: {folder}"
                    "Run this this script as a different user probably, or check permissions of the web_root."
                )
                logger.exception(err)
                raise PermissionError(err) from exc

    # endregion

    # Region File Cache and File List
    async def update_file_cache(self) -> None:
        """Update the file cache."""
        if self.s3:
            await s3_file_cache.get_all(self._app_config.s3.bucket)
        else:
            local_file_cache.refresh(get_app_paths().web_root)

    async def get_file_list(self) -> APFileList:
        """Gets the base url and the file cache."""
        await self.update_file_cache()

        base_url = self._app_config.s3.cdn_domain if self.s3 else self._app_config.inet_path

        file_list = (
            await s3_file_cache.get_all_list_str(self._app_config.s3.bucket)
            if self.s3
            else local_file_cache.get_all_str()
        )

        return APFileList(base_url=base_url.encoded_string(), files=file_list)

    # region Rendering

    async def write_health_s3(self) -> None:
        """Write the health api to s3."""
        health_api_response = health.get_health()
        await self.renderer.write_health_s3(health_api_response)

    # endregion



"""Main function, for running adhoc."""

import argparse
from pathlib import Path

from . import run_ap_adhoc
from .constants import DEFAULT_INSTANCE_PATH
from .utils import logger as ap_logger
from .utils.log_messages import log_intro


def main() -> None:
    """Main function for CLI."""
    ap_logger.setup_logger(app=None)  # Setup logger with defaults defined in config module

    logger = ap_logger.get_logger(__name__)
    log_intro("adhoc", logger)

    parser = argparse.ArgumentParser(description="Archivepodcast.")
    parser.add_argument(
        "--instance-path",
        type=str,
        default="",
        help="Path to the instance directory.",
    )

    args = parser.parse_args()

    instance_path = Path(args.instance_path) if args.instance_path else None

    if not instance_path:
        msg = f"Instance path not provided, using default: {DEFAULT_INSTANCE_PATH}"
        logger.info(msg)
        instance_path = DEFAULT_INSTANCE_PATH  # pragma: no cover # This avoids issues in PyTest
        if not instance_path.exists():
            msg = f"Instance path ({instance_path}) does not exist, not creating it for safety."
            raise FileNotFoundError(msg)

    run_ap_adhoc(instance_path=instance_path)


if __name__ == "__main__":
    main()  # pragma: no cover



"""Logging configuration for archivepodcast."""

import logging
import os
from logging import StreamHandler
from logging.handlers import RotatingFileHandler
from pathlib import Path
from typing import Any, Self, cast

from flask import Flask
from pydantic import BaseModel, field_validator, model_validator
from rich.console import Console
from rich.highlighter import NullHighlighter
from rich.logging import RichHandler
from rich.theme import Theme

DESIRED_LEVEL_NAME_LEN = 5
DESIRED_NAME_LEN = 16
DESIRED_THREAD_NAME_LEN = 13


LOG_LEVELS = [
    "TRACE",
    "DEBUG",
    "INFO",
    "WARNING",
    "ERROR",
    "CRITICAL",
]  # Valid str logging levels.


MIN_LOG_LEVEL_INT = 0
MAX_LOG_LEVEL_INT = 50


class LoggingConf(BaseModel):
    """Logging configuration definition."""

    level: str | int = "INFO"
    path: Path | None = None

    @model_validator(mode="after")
    def validate_vars(self) -> Self:
        """Validate the logging level."""
        if isinstance(self.level, int):
            if self.level < MIN_LOG_LEVEL_INT or self.level > MAX_LOG_LEVEL_INT:
                msg = (
                    f"Invalid logging level {self.level}, must be between {MIN_LOG_LEVEL_INT} and {MAX_LOG_LEVEL_INT}."
                )
                logger.warning(msg)
                logger.warning("Defaulting logging level to 'INFO'.")
                self.level = "INFO"
        else:
            self.level = self.level.strip().upper()
            if self.level not in LOG_LEVELS:
                msg = f"Invalid logging level '{self.level}', must be one of {', '.join(LOG_LEVELS)}"
                logger.warning(msg)
                logger.warning("Defaulting logging level to 'INFO'.")
                self.level = "INFO"

        return self

    @field_validator("path", mode="before")
    def set_path(cls, value: str | None) -> Path | None:  # noqa: N805 # ???
        """Set the path to a slugified version."""
        if value is None:
            return None

        if isinstance(value, str):
            value = value.strip()

        if value == "":
            return None

        return Path(value)

    def setup_verbosity_cli(self, verbosity: int) -> None:
        """Setup the logger from verbosity count from CLI."""
        if verbosity >= 2:  # noqa: PLR2004 Magic number makes sense
            self.level = TRACE_LEVEL_NUM
        elif verbosity == 1:
            self.level = logging.DEBUG
        else:
            self.level = logging.INFO


# This is the logging message format that I like.
# LOG_FORMAT = "%(asctime)s:%(levelname)s:%(name)s:%(message)s"   # noqa: ERA001
SIMPLE_LOG_FORMAT = "%(levelname)s:%(message)s"
SIMPLE_LOG_FORMAT_DEBUG = "%(levelname)s:%(name)s:%(message)s"
TRACE_LEVEL_NUM = 5


class CustomLogger(logging.Logger):
    """Custom logger to appease mypy."""

    def trace(self, message: object, *args: Any, **kws: Any) -> None:  # noqa: ANN401
        """Create logger level for trace."""
        if self.isEnabledFor(TRACE_LEVEL_NUM):
            # Yes, logger takes its '*args' as 'args'.
            self._log(TRACE_LEVEL_NUM, message, args, **kws)


logging.addLevelName(TRACE_LEVEL_NUM, "TRACE")
logging.setLoggerClass(CustomLogger)

# This is where we log to in this module, following the standard of every module.
# I don't use the function so we can have this at the top
logger = cast("CustomLogger", logging.getLogger(__name__))

# In flask the root logger doesn't have any handlers, its all in app.logger
# root_logger : root,
# app.logger  : root, archivepodcast,
# logger      : root, archivepodcast, archivepodcast.module_name,
# The issue is that waitress, werkzeug (any any other modules that log) will log separately.
# The aim is, remove the default handler from the flask App and create one on the root logger to apply config to all.


# Pass in the whole app object to make it obvious we are configuring the logger object within the app object.
def setup_logger(
    app: Flask | None,
    logging_conf: LoggingConf | None = None,
    in_logger: logging.Logger | None = None,
) -> None:
    """Configure logging for the application.

    Args:
        app: The Flask application instance
        logging_conf: Logging configuration object
        in_logger: Optional logger instance to configure (mainly for testing)
    """
    if logging_conf is None:
        logging_conf = LoggingConf()

    if not in_logger:  # in_logger should only exist when testing with PyTest.
        in_logger = logging.getLogger()  # Get the root logger

    # The root logger has no handlers initially in flask, app.logger does though.
    if app:
        app.logger.handlers.clear()  # Remove the Flask default handlers

    if not running_in_serverless_environment():
        in_logger.handlers.clear()

    # If the logger doesn't have a console handler (root logger doesn't by default)
    if (not any(isinstance(handler, (RichHandler, StreamHandler)) for handler in in_logger.handlers)) and (
        not running_in_serverless_environment()  # Serverless should have their own handler
    ):
        _add_console_handler(logging_conf, in_logger)

    _set_log_level(in_logger, logging_conf.level)

    # If we are logging to a file
    if not _has_file_handler(in_logger) and logging_conf.path is not None:
        _add_file_handler(in_logger, logging_conf.path)

    # Configure modules that are external and have their own loggers
    logging.getLogger("waitress").setLevel(logging.INFO)  # Prod web server, info has useful info.
    logging.getLogger("werkzeug").setLevel(logging.DEBUG)  # Only will be used in dev, debug logs incoming requests.
    logging.getLogger("urllib3").setLevel(logging.WARNING)  # Bit noisy when set to info, used by requests module.
    logging.getLogger("botocore").setLevel(logging.WARNING)  # Can be noisy
    logging.getLogger("boto3").setLevel(logging.WARNING)  # Can be noisy
    logging.getLogger("s3transfer").setLevel(logging.WARNING)  # Can be noisy
    logging.getLogger("aiobotocore").setLevel(logging.INFO)  # Can be noisy
    logging.getLogger("asyncio").setLevel(logging.INFO)  # Can be noisy

    logger.debug("Logger configuration set!")


def get_logger(name: str) -> CustomLogger:
    """Get a logger with the name provided."""
    return cast("CustomLogger", logging.getLogger(name))


def _has_file_handler(in_logger: logging.Logger) -> bool:
    """Check if logger has a file handler."""
    return any(isinstance(handler, logging.FileHandler) for handler in in_logger.handlers)


def _add_console_handler(
    settings: LoggingConf,
    in_logger: logging.Logger,
) -> None:
    """Add a console handler to the logger."""
    if not force_simple_logger():
        console = Console(theme=Theme({"logging.level.trace": "dim"}))
        rich_handler = RichHandler(
            console=console,
            show_time=False,
            rich_tracebacks=True,
            highlighter=NullHighlighter(),
        )
        in_logger.addHandler(rich_handler)
    else:
        console_handler = StreamHandler()
        if _get_log_level_int(settings.level) <= logging.DEBUG:
            formatter = logging.Formatter(SIMPLE_LOG_FORMAT_DEBUG)
        else:
            formatter = logging.Formatter(SIMPLE_LOG_FORMAT)

        console_handler.setFormatter(formatter)
        in_logger.addHandler(console_handler)


def _get_log_level_int(level: str | int) -> int:
    """Get the log level as an int."""
    if isinstance(level, int):
        return level

    level = level.upper()
    if level == "TRACE":
        return TRACE_LEVEL_NUM
    return getattr(logging, level, logging.INFO)


def _set_log_level(in_logger: logging.Logger, log_level: int | str) -> None:
    """Set the log level of the logger."""
    if isinstance(log_level, str):
        log_level = log_level.upper()
        if log_level not in LOG_LEVELS:
            in_logger.setLevel("INFO")
            logger.warning(
                "❗ Invalid logging level: %s, defaulting to INFO",
                log_level,
            )
        else:
            in_logger.setLevel(log_level)
            logger.debug("Showing log level: INFO and above")
            logger.debug("Showing log level: DEBUG")
            logger.trace("Showing log level: TRACE")
    else:
        in_logger.setLevel(log_level)


def _add_file_handler(in_logger: logging.Logger, log_path: Path | str) -> None:
    """Add a file handler to the logger."""
    if not isinstance(log_path, Path):
        log_path = Path(log_path)

    try:
        log_path.parent.mkdir(parents=True, exist_ok=True)
        file_handler = RotatingFileHandler(log_path, maxBytes=1000000, backupCount=5)
    except IsADirectoryError as exc:
        err = "You are trying to log to a directory, try a file"
        raise IsADirectoryError(err) from exc
    except PermissionError as exc:
        err = f"The user running this does not have access to the file: {log_path}"
        raise PermissionError(err) from exc

    formatter = logging.Formatter(SIMPLE_LOG_FORMAT)
    file_handler.setFormatter(formatter)
    in_logger.addHandler(file_handler)
    logger.info("Logging to file: %s", log_path)


def running_in_serverless_environment() -> bool:
    """Check if the application is running in a serverless environment."""
    return os.getenv("AWS_LAMBDA_FUNCTION_NAME") is not None


def force_simple_logger() -> bool:
    """Check if the application is running in a serverless environment."""
    return running_in_serverless_environment() or (os.getenv("AP_SIMPLE_LOGGING", "").lower() in ["1", "true"])



"""Configuration management for archivepodcast."""

import json
from datetime import UTC, datetime
from pathlib import Path
from typing import Literal, Self

from pydantic import BaseModel, HttpUrl, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from .utils.logger import LoggingConf, get_logger

# Logging should be all done at INFO level or higher as the log level hasn't been set yet
# Modules should all setup logging like this so the log messages include the modules name.
logger = get_logger(__name__)


class AppWebPageConfig(BaseModel):
    """App Web Page Config Object."""

    title: str = "Podcast Archive"
    description: str = (
        "Podcast archive, generated by archivepodcast, available at https://github.com/kism/archivepodcast"
    )
    contact: str = "archivepodcast@localhost"


class AppS3Config(BaseModel):
    """App S3 Config Object."""

    cdn_domain: HttpUrl = HttpUrl("https://example.com/public_url_of_s3_bucket/")
    api_url: HttpUrl | None = None
    bucket: str = ""
    region: str = ""
    access_key_id: str = ""
    secret_access_key: str = ""

    @field_validator("api_url", mode="before")
    def validate_api_url(cls, v: str) -> str | None:  # noqa: N805
        """Validate api_url to allow empty string to be None."""
        if v == "":
            return None

        return v


class AppConfig(BaseModel):
    """App Config Object."""

    web_page: AppWebPageConfig = AppWebPageConfig()
    inet_path: HttpUrl = HttpUrl("http://localhost:5100/")
    storage_backend: Literal["local", "s3"] = "local"
    s3: AppS3Config = AppS3Config()


class PodcastConfig(BaseModel):
    """Podcast Config Object."""

    url: HttpUrl = HttpUrl("https://example.com/podcast/feed.xml")
    new_name: str = ""
    name_one_word: str = ""
    description: str = ""
    live: bool = True
    contact_email: str = ""


class FlaskConfig(BaseModel):
    """Flask Config Object."""

    TESTING: bool = False


class ArchivePodcastConfig(BaseSettings):
    """Config Object."""

    model_config = SettingsConfigDict(
        extra="ignore",
        env_file=".env",
        env_prefix="AP_",
        case_sensitive=False,
        env_file_encoding="utf-8",
        env_nested_delimiter="__",
    )

    app: AppConfig = AppConfig()
    podcasts: list[PodcastConfig] = [PodcastConfig()]
    logging: LoggingConf = LoggingConf()
    flask: FlaskConfig = FlaskConfig()

    def write_config(self, config_path: Path) -> None:
        """Write the current settings to a JSON file."""
        config_path.parent.mkdir(parents=True, exist_ok=True)

        config_data = json.loads(self.model_dump_json())

        if not config_path.exists():
            logger.warning("Writing fresh config file at %s", config_path.absolute())
            config_path.touch()
            existing_data = config_data
        else:
            with config_path.open("r") as f:
                existing_data = json.load(f)

        new_file_content_str = json.dumps(config_data)

        if existing_data != config_data:  # The new object will be valid, so we back up the old one
            time_str = datetime.now(tz=UTC).strftime("%Y-%m-%d_%H%M%S")
            config_backup_dir = config_path.parent / "config_backups"
            config_backup_dir.mkdir(parents=True, exist_ok=True)
            backup_file = config_backup_dir / f"{config_path.stem}_{time_str}{config_path.suffix}.bak"
            logger.warning(
                "Validation has changed the config file, backing up the old one to %s",
                backup_file,
            )
            with backup_file.open("w") as f:
                f.write(json.dumps(existing_data))

        with config_path.open("w") as f:
            f.write(new_file_content_str)

        config_path_json = config_path.with_suffix(".json")
        logger.info("Writing config to %s", config_path_json)
        with config_path_json.open("w") as f:
            f.write(self.model_dump_json(indent=2, exclude_none=False))

        logger.info("Config write complete")

    @classmethod
    def force_load_config_file(cls, config_path: Path) -> Self:
        """Load the configuration file. File contents takes precedence over env vars."""
        if not config_path.exists():
            logger.warning(
                "Config file %s does not exist, loading defaults",
                config_path.absolute(),
            )
            return cls()

        logger.info("Loading config from %s", config_path.absolute())
        with config_path.open("r") as f:
            config = json.load(f)

        return cls(**config)

    def post_validate(self) -> None:
        """Post validation function."""
        for i, podcast in enumerate(self.podcasts):
            if podcast.name_one_word == "":
                msg = f"Please fill in the podcast details on entry {i}\n"
                msg += podcast.model_dump_json()
                raise ValueError(msg)



"""Log messages for ArchivePodcast."""

import logging
from datetime import datetime

from aiohttp import ClientError

from archivepodcast.constants import OUR_TIMEZONE
from archivepodcast.version import __version__


def log_time(logger: logging.Logger) -> None:
    """Log the current time."""
    time = datetime.now(tz=OUR_TIMEZONE)
    time_str_nice = time.strftime("%Y-%m-%d %H:%M:%S %Z")
    logger.info("Current time: %s", time_str_nice)


def log_intro(mode: str, logger: logging.Logger) -> None:
    """Log introductory information."""
    logger.info("ArchivePodcast version: %s, starting in %s mode.", __version__, mode)
    log_time(logger)


def log_aiohttp_exception(
    feed: str,
    url: str,
    exception: ClientError,
    logger: logging.Logger,
) -> None:
    """Log an aiohttp exception with details."""
    logger.error("[%s] download error for %s: %s", feed, url, type(exception).__name__)



"""Helper utilities for archivepodcast."""

from datetime import UTC, datetime
from typing import TYPE_CHECKING

from aiobotocore.session import get_session
from pydantic import BaseModel

from archivepodcast.instances.config import get_ap_config_s3_client

from .logger import get_logger

logger = get_logger(__name__)

if TYPE_CHECKING:
    from types_aiobotocore_s3.type_defs import ObjectTypeDef  # pragma: no cover
else:
    ObjectTypeDef = object

MAX_CACHE_AGE = 120


class S3File(BaseModel):
    """Model representing a S3 file in the cache."""

    key: str
    size: int


class S3FileCache(BaseModel):
    """Model representing a cache of S3 files."""

    _last_cache_time: datetime | None = None

    _files: list[ObjectTypeDef] = []

    async def get_all(self, bucket: str) -> list[ObjectTypeDef]:
        """List all objects in an S3 bucket using pagination."""
        if self._last_cache_time:
            age = (datetime.now(tz=UTC) - self._last_cache_time).total_seconds()
            logger.trace("S3 Cache hit! Age: %.2f seconds", age)
            if age < MAX_CACHE_AGE:
                return self._files

        logger.debug("Fetching object list from S3, no cache available")

        s3_config = get_ap_config_s3_client()

        session = get_session()
        async with session.create_client("s3", **s3_config.model_dump()) as s3_client:
            paginator = s3_client.get_paginator("list_objects_v2")
            page_iterator = paginator.paginate(Bucket=bucket)

            all_objects: list[ObjectTypeDef] = []
            async for page in page_iterator:
                if "Contents" in page:
                    all_objects.extend(page["Contents"])

        self._files = all_objects
        self._last_cache_time = datetime.now(tz=UTC)
        return all_objects

    async def get_all_list_str(self, bucket: str) -> list[str]:
        """Get all S3 file keys as strings."""
        s3_files = await self.get_all(bucket)
        return [s3_file["Key"] for s3_file in s3_files]

    def add_file(self, s3_file: S3File) -> None:
        """Append a new S3 file to the cache.

        Args:
            s3_file: S3File object to append
        """
        self._files.append({"Key": s3_file.key, "Size": s3_file.size})

    def check_file_exists(self, key: str, size: int | None = None) -> bool:
        """Check if a file exists in the cache."""
        matching_file: ObjectTypeDef | None = None
        for file in self._files:
            if file["Key"] == key:
                matching_file = file
                break

        if size is not None:
            return any(file["Key"] == key and file["Size"] == size for file in self._files)

        return matching_file is not None



"""Constants for the ArchivePodcast application."""

from datetime import UTC, datetime
from pathlib import Path

OUR_TIMEZONE = datetime.now().astimezone().tzinfo or UTC
APP_DIRECTORY = Path(__file__).parent
DEFAULT_INSTANCE_PATH = Path.cwd() / "instance"



"""Health monitoring for archivepodcast components."""

import contextlib
import datetime
from typing import TYPE_CHECKING, Self

from lxml import etree
from psutil import Process
from pydantic import BaseModel

from archivepodcast.utils.logger import get_logger
from archivepodcast.version import __version__

if TYPE_CHECKING:
    from archivepodcast.archiver import PodcastArchiver  # pragma: no cover
    from archivepodcast.config import AppConfig  # pragma: no cover
else:
    PodcastArchiver = object
    AppConfig = object

logger = get_logger(__name__)

PODCAST_DATE_FORMATS = ["%a, %d %b %Y %H:%M:%S %z", "%a, %d %b %Y %H:%M:%S GMT"]


PROCESS = Process()


class EpisodeInfo(BaseModel):
    """Episode Info Model."""

    title: str = "Unknown"
    pubdate: int = 0


class PodcastHealth(BaseModel):
    """Health status for an individual podcast."""

    latest_episode: EpisodeInfo = EpisodeInfo()
    rss_available: bool = False
    rss_fetching_live: bool = False
    last_fetched: int = 0
    healthy_download: bool | None = None
    healthy_feed: bool = False
    episode_count: int = 0

    def update_episode_info(self, tree: etree._ElementTree | etree._Element | None = None) -> None:
        """Update the latest episode info."""
        logger.trace("Updating podcast episode info")
        new_latest_episode: EpisodeInfo = EpisodeInfo()
        new_episode_count: int = 0

        try:
            if tree is not None:
                if len(tree.xpath("//item")) == 0:
                    logger.warning("No episodes found in feed")
                    self.latest_episode = new_latest_episode
                    self.episode_count = new_episode_count
                    return

                latest_episode = tree.xpath("//item")[0]

                # If we have the title, use it
                with contextlib.suppress(IndexError):
                    new_latest_episode.title = latest_episode.xpath("title")[0].text

                # If we have the description, use it
                with contextlib.suppress(IndexError):
                    new_episode_count = len(tree.xpath("//item"))

                # If we have the pubDate, try to parse it
                if len(latest_episode.xpath("pubDate")) > 0 and latest_episode.xpath("pubDate")[0].text:
                    pod_pubdate = str(latest_episode.xpath("pubDate")[0].text) or "1970-01-01 00:00:00"
                    found_pubdate = False
                    for podcast_date_format in PODCAST_DATE_FORMATS:
                        try:
                            new_latest_episode.pubdate = int(
                                datetime.datetime.strptime(pod_pubdate, podcast_date_format)
                                .replace(tzinfo=datetime.UTC)
                                .timestamp()
                            )
                            found_pubdate = True
                            break
                        except ValueError:
                            pass
                    if not found_pubdate:
                        logger.error("Unable to parse pubDate: %s", pod_pubdate)
        except Exception:  # pragma: no cover # Just to be safe
            logger.exception("Error parsing podcast episode info")  # pragma: no cover # Just to be safe

        self.latest_episode = new_latest_episode
        self.episode_count = new_episode_count


class WebpageHealth(BaseModel):
    """Health status for an individual webpage."""

    last_render: int = int(datetime.datetime.now(tz=datetime.UTC).timestamp())


class Host(BaseModel):
    """Info about a host config entry."""

    host_type: str = ""
    url: str = ""  # Don't use HttpUrl since we aren't validating


class HostingInfo(BaseModel):
    """Info about how the site is hosted."""

    backend: Host = Host()
    frontend: Host = Host()

    @classmethod
    def load_from_config(cls, app_config: AppConfig) -> Self:
        """Load HostType from AppConfig."""
        frontend_host_type = "Flask"
        if app_config.storage_backend == "s3" and app_config.inet_path == app_config.s3.cdn_domain:
            frontend_host_type = "s3"

        backend_host_type = "s3"
        if app_config.storage_backend == "local":
            backend_host_type = "Flask"

        backend = Host(
            host_type=backend_host_type,
            url=app_config.s3.cdn_domain.encoded_string(),
        )
        frontend = Host(
            host_type=frontend_host_type,
            url=app_config.inet_path.encoded_string(),
        )
        return cls(backend=backend, frontend=frontend)


class CoreHealth(BaseModel):
    """Core application health status."""

    alive: bool = True
    last_run: int = 0
    about_page_exists: bool = False
    last_startup: int = int(datetime.datetime.now(tz=datetime.UTC).timestamp())
    currently_rendering: bool = False
    currently_loading_config: bool = False
    memory_mb: float = -0.0
    debug: bool = False


class PodcastArchiverHealthAPI(BaseModel):
    """Podcast Archiver Health API Model."""

    core: CoreHealth
    podcasts: dict[str, PodcastHealth]
    templates: dict[str, WebpageHealth]
    version: str
    assets: dict[str, str]
    host_info: HostingInfo


class PodcastArchiverHealth:
    """Overall health monitoring for PodcastArchiver."""

    def __init__(self) -> None:
        """Initialise the Podcast Archiver Health object."""
        self._core: CoreHealth = CoreHealth()
        self._podcasts: dict[str, PodcastHealth] = {}
        self._templates: dict[str, WebpageHealth] = {}
        self._assets: dict[str, str] = {}
        self._version: str = __version__
        self._host_info: HostingInfo = HostingInfo()

    def get_health(self) -> PodcastArchiverHealthAPI:
        """Return the health."""
        if PROCESS is not None:
            self._core.memory_mb = PROCESS.memory_info().rss / (1024 * 1024)

        return PodcastArchiverHealthAPI(
            core=self._core,
            podcasts=self._podcasts,
            templates=self._templates,
            assets=self._assets,
            version=self._version,
            host_info=self._host_info,
        )

    def set_asset(self, path: str, mime: str) -> None:
        """Set an asset."""
        self._assets[path] = mime

    def update_template_status(self, webpage: str, **kwargs: bool | str | int | None) -> None:
        """Update the webpage."""
        if webpage not in self._templates:
            self._templates[webpage] = WebpageHealth()

        for key, value in kwargs.items():
            if value is not None and hasattr(self._templates[webpage], key):
                setattr(self._templates[webpage], key, value)

    def update_podcast_status(self, podcast: str, **kwargs: bool | str | int | None) -> None:
        """Update the podcast."""
        logger.trace("Updating podcast health for %s: %s", podcast, kwargs)
        if podcast not in self._podcasts:
            self._podcasts[podcast] = PodcastHealth()

        for key, value in kwargs.items():
            if value is not None and hasattr(self._podcasts[podcast], key):
                setattr(self._podcasts[podcast], key, value)

    def update_podcast_episode_info(self, podcast: str, tree: etree._ElementTree | etree._Element) -> None:
        """Update the podcast episode info."""
        logger.trace("Updating podcast episode info for %s", podcast)
        if podcast not in self._podcasts:
            self._podcasts[podcast] = PodcastHealth()

        self._podcasts[podcast].update_episode_info(tree)

    def update_core_status(self, **kwargs: bool | str | int | None) -> None:
        """Update the core."""
        valid_attrs = self._core.model_dump().keys()
        for key, value in kwargs.items():
            if value is not None and key in valid_attrs:
                setattr(self._core, key, value)

    def currently_rendering(
        self,
    ) -> bool:
        """Set currently rendering status."""
        return self._core.currently_rendering

    def currently_loading_config(
        self,
    ) -> bool:
        """Set currently loading config status."""
        return self._core.currently_loading_config

    def set_host_info(self, app_config: AppConfig) -> None:
        """Set the hosting info from AppConfig."""
        self._host_info = HostingInfo.load_from_config(app_config)



"""Webpage blueprints for ArchivePodcast."""

from http import HTTPStatus

from flask import Blueprint, Response

from archivepodcast.instances.podcast_archiver import (
    generate_404,
    get_about_page_exists,
    send_ap_cached_webpage,
)
from archivepodcast.utils.logger import get_logger

logger = get_logger(__name__)


bp = Blueprint("webpages", __name__)


@bp.route("/")  # type: ignore[untyped-decorator]
def home() -> Response:
    """Flask Home.

    If you are serving static files with s3 or nginx, ensure that / redirects to /index.html,
    """
    return Response(
        "Redirecting to /index.html", status=HTTPStatus.TEMPORARY_REDIRECT, headers={"Location": "/index.html"}
    )


@bp.route("/index.html")  # type: ignore[untyped-decorator]
def home_index() -> Response:
    """Flask Home."""
    return send_ap_cached_webpage("index.html")


@bp.route("/guide.html")  # type: ignore[untyped-decorator]
def home_guide() -> Response:
    """Podcast app guide."""
    return send_ap_cached_webpage("guide.html")


@bp.route("/webplayer.html")  # type: ignore[untyped-decorator]
def home_web_player() -> Response:
    """Podcast app guide."""
    return send_ap_cached_webpage("webplayer.html")


@bp.route("/about.html")  # type: ignore[untyped-decorator]
def home_about() -> Response:
    """Flask Home, s3 backup compatible."""
    if get_about_page_exists():
        return send_ap_cached_webpage("about.html")

    return generate_404()


@bp.route("/health")  # type: ignore[untyped-decorator]
@bp.route("/health.html")  # type: ignore[untyped-decorator]
def health() -> Response:
    """Health check."""
    return send_ap_cached_webpage("health.html")


@bp.route("/filelist.html")  # type: ignore[untyped-decorator]
def home_filelist() -> Response:
    """Serve Filelist."""
    return send_ap_cached_webpage("filelist.html")
